#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* parse permissions in symbolic form (adapted from V7 chmod) */
#define	USER	05700	/* user's bits */
#define	GROUP	02070	/* group's bits */
#define	OTHER	00007	/* other's bits */
/* V7 defines ALL 01777 and a comment highlights the fact that s[ug]id and stick bits
 * are not included, but POSIX says 'a' should be equivalent to 'ugo' and current
 * chmods agree
 */
#define	ALL	(USER|GROUP|OTHER)

#define	READ	00444	/* read permit */
#define	WRITE	00222	/* write permit */
#define	EXEC	00111	/* exec permit */
#define	SETID	06000	/* set[ug]id */
#define	STICKY	01000	/* sticky bit */

mode_t parse_perms(char *p, mode_t tm)
{
	mode_t cumask = 0;
	int is_dir = S_ISDIR(tm);
	tm = tm & 07777;

	do {
		int i;
		unsigned int who, perm; 
		char op;

		/* first, a sequence of zero or more chars in [ugoa].  These
		 * determine which set(s) of permissions we need to change. We
		 * store this information in a 'who' bitmap: bit 2 for u, bit 1
		 * for g and bit 0 for o.
		 */
		who = 0;
		for (;;) {
			if (*p == 'u')
				who |= USER;
			else if (*p == 'g')
				who |= GROUP;
			else if (*p == 'o')
				who |= OTHER;
			else if (*p == 'a')
				who = ALL;
			else
				break;
			p++;
		}
		if (!who) {
			/* by default apply to all, but use the current umask
			 * (Note that the umask is not used otherwise)
			 */
			/* there is no way to read the current umask, other than
			 * setting it to something and get the previous one.
			 */
			cumask = umask(0);
			/* we don't need to reset it to its original value, since
			 * we are not going to create new files
			 */
			who = ALL & ~cumask;
		}
		/* now a sequence of one or more actions, where each action is
		 * an operation (one of =, + or -) followed by a set of
		 * permissions (zero or more of r, w, x, X, s, t, u, g, o).  We
		 * remember the operation in 'op' and collect the permissions
		 * in the 'perm' bitmap.
		 */
		op = -1; // if 'op' is still -1 after this cycle, we
			 // didn't see any action
		while (*p != '\0' && strchr("=+-", *p)) {
			op = *p++;
			perm = 0;
			if (*p != '\0' && strchr("ugo", *p)) {
				switch (*p) {
				case 'u': perm = (tm & USER) >> 6; break;
				case 'g': perm = (tm & GROUP) >> 3; break;
				case 'o': perm = (tm & OTHER); break;
				}
				perm &= (READ|WRITE|EXEC);
				perm |= (perm << 6) | (perm << 3);
				p++;
			} else {
				while (*p != '\0' && !strchr("=+-,", *p)) {
					switch (*p) {
					case 'r': perm |= READ; break;
					case 'w': perm |= WRITE; break;
					case 'x': perm |= EXEC; break;
					case 'X': perm |= (is_dir || (tm & EXEC)) ? EXEC : 0; break;
					case 's': perm |= SETID; break;
					case 't': perm |= STICKY; break;
					default:
						fprintf(stderr, "syntax error\n");
						exit(1);
					}
					p++;
				}
			}
			/* now apply 'perm' to 'who' according to the selected
			 * operation (as stored in 'op')
			 */
			switch (op) {
			case '+':
				tm |= perm & who;
				break;
			case '-':
				tm &= ~(perm & who);
				break;
			case '=':
				tm &= ~who;
				tm |= perm & who;
				break;
			}
		}
		if (op == -1) {
			fprintf(stderr, "syntax error\n");
			exit(1);
		}
	} while (*p++ == ',');
	// return the computed new-mode
	return tm;
}

int main(int argc, char* argv[])
{
	mode_t tm = 0;
	int errors;
	int i;
	int needstat;
	struct stat statbuf;

	if (argc < 3) {
		fprintf(stderr, "usage: chmod <mode> <file>...\n");
		exit(1);
	}

	// to process symbolic input in general, we need to stat() every file
	// to get its current permissions.  We can avoid that if the input is
	// numeric.
	needstat = 1;
	if (isdigit(argv[1][0])) {
		tm = strtol(argv[1], NULL, 8);
		needstat = 0;
	}

	errors = 0;
	for (i = 2; i < argc; i++) {
		if (needstat) {
			if (stat(argv[i], &statbuf) < 0) {
				perror(argv[i]);
				continue;
			}
			tm = parse_perms(argv[1], statbuf.st_mode);
		}
		if (chmod(argv[i], tm) < 0) {
			perror(argv[i]);
			errors++;
		}
	}
	return errors;
}
