#include <sys/types.h>
#include <dirent.h>
#include <inttypes.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>

/*
 * ls that understands symbolic links
 *
 * To show symbolic links correctly we need to use lstat() instead of stat(),
 * since the latter one would follow the link and return the inode of the
 * pointed-to entity (if any).
 */

#define MAX_LINK 1024

void pretty_print(struct stat* s);

int main(int argc, char* argv[])
{
	DIR *dir;
	struct dirent* ent;
	int all = 0, inodes = 0, longl = 0;
	struct stat statbuf;
	char *dirname = ".";
	char linkbuf[MAX_LINK + 1];
	int n;

	if (argc > 1)
		if (argv[1][0] == '-') {
			char* opt = &argv[1][1];
			while (*opt) {
				switch (*opt) {
				case 'a':
					all = 1;
					break;
				case 'i':
					inodes = 1;
					break;
				case 'l':
					longl = 1;
					break;
				default:
					fprintf(stderr, "Unknown option: %c\n", *opt);
					break;
				}
				opt++;
			}
			argc--;
			argv++;
		}

	if (argc > 1)
		dirname = argv[1];

	if ( (dir = opendir(dirname)) == NULL) {
		perror(dirname);
		exit(1);
	}

	if (longl && chdir(dirname) < 0) {
		perror(dirname);
		exit(1);
	}

	while ( (ent = readdir(dir)) ) {

		if (ent->d_name[0] == '.' && !all)
			continue;

		if (inodes)
			printf("%10llu ", (unsigned long long)ent->d_ino);

		if (longl) {
			if (lstat(ent->d_name, &statbuf) < 0) {
				perror(ent->d_name);
			} else {
				pretty_print(&statbuf);
			}
		}
		printf("%s", ent->d_name);

		if (longl && S_ISLNK(statbuf.st_mode)) {
			if ( (n = readlink(ent->d_name, linkbuf, MAX_LINK)) < 0) {
				perror(ent->d_name);
			} else {
				linkbuf[n] = '\0';
				printf(" -> %s", linkbuf);
			}
		}

		printf("\n");
	}

	return 0;
}

char* pretty_print_mode(mode_t m);
char* pretty_print_owner(uid_t o);
char *pretty_print_group(gid_t g);
char* pretty_print_mtime(time_t *t);

void pretty_print(struct stat* s)
{
	if (S_ISCHR(s->st_mode) || S_ISBLK(s->st_mode)) {
		printf("%10s %3ju %10s %10s %3d, %3d %s ",
			pretty_print_mode(s->st_mode),
			(uintmax_t)s->st_nlink,
			pretty_print_owner(s->st_uid),
			pretty_print_group(s->st_gid),
			((unsigned short)s->st_rdev & 0xff00) >> 8,
			((unsigned short)s->st_rdev & 0x00ff),
			pretty_print_mtime(&s->st_mtime));
	} else {
		printf("%10s %3ju %10s %10s %8jd %s ",
			pretty_print_mode(s->st_mode),
			(uintmax_t)s->st_nlink,
			pretty_print_owner(s->st_uid),
			pretty_print_group(s->st_gid),
			(intmax_t)s->st_size,
			pretty_print_mtime(&s->st_mtime));
	}
}

char* pretty_print_mode(mode_t m)
{
	static char mode[11];
	int i;
	char modes[] = { 'r', 'w', 'x' };
	mode_t m1;

	switch (m & S_IFMT) {
	case S_IFLNK:
		mode[0] = 'l';
		break;
	case S_IFREG:
		mode[0] = '-';
		break;
	case S_IFBLK:
		mode[0] = 'b';
		break;
	case S_IFDIR:
		mode[0] = 'd';
		break;
	case S_IFCHR:
		mode[0] = 'c';
		break;
	default:
		mode[0] = '?';
		break;
	}
	m1 = m;
	m1 &= (S_IRWXU | S_IRWXG | S_IRWXO);
	for (i = 9; i > 0; i--) {
		if (m1 & 1) {
			mode[i] = modes[(i - 1) % 3];
		} else {
			mode[i] = '-';
		}
		m1 >>= 1;
	}
	mode[10] = '\0';
	if (m & S_ISUID)
		mode[3] = (mode[3] == 'x') ? 's' : 'S';

	if (m & S_ISGID)
		mode[6] = (mode[6] == 'x') ? 's' : 'S';

	if (m & S_ISVTX)
		mode[9] = (mode[9] == 'x') ? 't' : 'T';

	return mode;
}

#define MAX_LINE 80

char* pretty_print_owner(uid_t u)
{
	static char owner[11];
	struct passwd *pw;

	snprintf(owner, 11, "%d", u);

	if ( (pw = getpwuid(u)) != NULL )
		return pw->pw_name;

	return owner;
}

char* pretty_print_group(gid_t g)
{
	static char group[11];
	struct group *gr;

	snprintf(group, 11, "%d", g);

	if ( (gr = getgrgid(g)) != NULL )
		return gr->gr_name;

	return group;
}

char* pretty_print_mtime(time_t *t)
{
	static char mytime[14];
	strftime(mytime, 13, "%b %e %H:%M", localtime(t));
	return mytime;
}
