#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* This shell keeps track of the return value of commands.  It prints a message
 * for programs that terminate abnormally (i.e., killed by a signal). The value
 * returned by programs that terminate normally is available in the special
 * internal variable $?. We also add the $$ special variable, that contains the
 * pid of the shell process.
 */

#define MAX_LINE 1024
#define MAX_WORDS 100
#define MAX_ARGS  100

int interactive;
int error;
sig_t inth, quith;
int lastresult;

pid_t shellpid;

int interp();

int main()
{
	setenv("IFS", " \t\n", 1);

	setbuf(stdin, NULL);
	shellpid = getpid();

	interactive = isatty(0);
	error = 0;
	inth = SIG_DFL;
	quith = SIG_DFL;
	if (interactive) {
		inth = signal(SIGINT, SIG_IGN);
		quith = signal(SIGQUIT, SIG_IGN);
	}
	lastresult = 0;

	return interp();
}

typedef struct {
	char *w;
	int  type;
#define W_NORML	0
#define	W_REDIR	1
#define	W_ASSGN	2
	int quot;
} word_t;

#define QUOTE	0x80

int getwords(char *buf, word_t words[]);
int buildargv(char *argv[], word_t words[], int nwords);
void getredirs(word_t words[], int nwords);
int redirect(word_t words[], int nwords);
int getcmd(char *buf);
void getvars(word_t words[], int nwords);
void assignvars(word_t words[], int nwords);
int expandall(char *buf, word_t words[], int nwords);
void quote(char *src);
void unquoteall(word_t words[], int nwords);
void docmd(char *buf);

typedef int (*builtin_t)(int argc, char *argv[]);

typedef struct {
	const char *name;
	builtin_t  builtin;
} builtin_des;

int builtin_cd(int argc, char *argv[]);
int builtin_umask(int argc, char *argv[]);
int builtin_unset(int argc, char *argv[]);
int builtin_colon(int argc, char *argv[]);
builtin_des builtins[] = {
	"cd",		builtin_cd,
	"umask",	builtin_umask,
	"unset",	builtin_unset,
	":",		builtin_colon,
	NULL, NULL
};
builtin_t findbuiltin(const char *name);

struct obuf {
	char *buf;
	int avail;
	int size;
};

#define oinit(o, b) do {			\
		typeof(o) o__ = o;		\
		(o__)->buf = b;			\
		(o__)->avail = sizeof(b);	\
		(o__)->size = sizeof(b);	\
	} while(0)

int oused(struct obuf *o)
{
	if (o->avail < 0)
		return o->size;
	return o->size - o->avail;
}

void oput(struct obuf *o, char c)
{
	if (o->avail <= 0) {
		o->avail = -1;
		return;
	}
	*o->buf++ = c;
	o->avail--;
}

void oputs(struct obuf *o, const char *s, int quote)
{
	while (*s != '\0')
		oput(o, *s++ ^ quote);
}

void oputf(struct obuf *o, int quote, const char *fmt, ...)
{
	char tmp[MAX_LINE];
	va_list va;
	int w;

	va_start(va, fmt);
	w = vsnprintf(tmp, MAX_LINE, fmt, va);
	va_end(va);
	if (w >= 0)
		oputs(o, tmp, quote);
}

int interp()
{
	char line[MAX_LINE];
	char *term;

	while (getcmd(line)) {
		quote(line);
		docmd(line);
	}
	return error ? error : lastresult;
}

void docmd(char *buf)
{
	word_t words[MAX_WORDS];
	char *c_argv[MAX_ARGS + 1];
	int nwords;
	int c_argc;
	builtin_t builtin;

	nwords = getwords(buf, words);
	if (!nwords || error)
		return;
	getredirs(words, nwords);
	getvars(words, nwords);
	nwords = expandall(buf, words, nwords);
	c_argc = buildargv(c_argv, words, nwords);
	builtin = findbuiltin(c_argv[0]);
	unquoteall(words, nwords);

	if (builtin) {
		assignvars(words, nwords);
		lastresult = (*builtin)(c_argc, c_argv);
		return;
	}

	if (fork()) {
		int status;

		wait(&status);
		if (WIFEXITED(status)) {
			lastresult = WEXITSTATUS(status);
		} else if (WIFSIGNALED(status)) {
			fprintf(stderr, "%s\n", strsignal(WTERMSIG(status)));
			lastresult = 0x80 | WTERMSIG(status);
		}
	} else {
		signal(SIGINT, inth);
		signal(SIGQUIT, quith);
		if (redirect(words, nwords) < 0)
			exit(1);
		if (c_argv[0] == NULL)
			exit(0);
		assignvars(words, nwords);
		if (builtin)
			exit((*builtin)(c_argc, c_argv));
		execvp(c_argv[0], c_argv);
		perror(c_argv[0]);
		exit(errno == ENOENT ? 127 : 126);
	}
}

int cleanupword(char *w)
{
	int i, j, q;

	q = 0;
	for (i = 0, j = 0; w[i] != '\0'; i++) {
		if (w[i] != '\'')
			w[j++] = w[i];
		else
			q = 1;
	}
	w[j] = '\0';
	return q;
}

void cleanupall(word_t words[], int nwords)
{
	int i;

	for (i = 0; i < nwords; i++) {
		word_t *nw = &words[i];

		nw->quot = cleanupword(nw->w);
	}
}

int getwords(char *buf, word_t words[])
{
	int nwords;
	int i;
	word_t *nw;

	nwords = 0;
	i = 0;
	for (;;) {
		while (isspace(buf[i]))
			i++;
		if (buf[i] == '\0')
			break;
		if (nwords >= MAX_WORDS) {
			fprintf(stderr, "too many words\n");
			error = 1;
			nwords = 0;
			break;
		}
		nw = &words[nwords];
		nw->w = &buf[i];
		nw->type = W_NORML;
		nwords++;
		while (buf[i] != '\0' && !isspace(buf[i]))
			i++;
		if (buf[i] == '\0')
			break;
		buf[i++] = '\0';
	}
	cleanupall(words, nwords);
	return nwords;
}

int buildargv(char *argv[], word_t words[], int nwords)
{
	int i, argc;

	argc = 0;
	for (i = 0; i < nwords && i < MAX_ARGS; i++) {
		word_t *nw = &words[i];
		if (nw->type != W_NORML)
			continue;
		argv[argc] = nw->w;
		argc++;
	}
	if (argc >= MAX_ARGS) {
		fprintf(stderr, "too many arguments\n");
		error = 1;
		return 0;
	}
	argv[argc] = NULL;
	return argc;
}

void getredirs(word_t words[], int nwords)
{
	int i;

	for (i = 0; i < nwords; i++) {
		word_t *nw = &words[i];
		if (nw->w[0] == '<' || nw->w[0] == '>') {
			nw->type = W_REDIR;
			if (nw->w[0] == '>' && nw->w[1] == '>')
				nw->w[0] = 'A';
		}
	}
}

int redirect(word_t words[], int nwords)
{
	int i, flags;
	char *fname;
	word_t *nw;

	for (i = 0; i < nwords; i++) {
		nw = &words[i];
		if (nw->type != W_REDIR)
			continue;
		fname = &nw->w[1];
		if (nw->w[0] == '<') {
			flags = O_RDONLY;
			close(0);
		} else if (nw->w[0] == 'A') {
			flags = O_WRONLY | O_CREAT | O_APPEND;
			fname++;
			close(1);
		} else {
			flags = O_WRONLY | O_CREAT | O_TRUNC;
			close(1);
		}
		if (open(fname, flags, 0666) < 0) {
			perror(fname);
			return -1;
		}
	}
	return 0;
}

int getcmd(char *buf)
{
	if (interactive) {
		error = 0;
		clearerr(stdin);
		fprintf(stderr, geteuid() ? "$ " : "# ");
	}
	if (error)
		return 0;
	if (fgets(buf, MAX_LINE, stdin) == NULL)
		return 0;
	if (strchr(buf, '\n') != NULL || feof(stdin))
		return 1;
	fprintf(stderr, "line too long\n");
	if (!interactive) {
		error = 1;
		return 0;
	}
	while (fgets(buf, MAX_LINE, stdin) != NULL &&
		strchr(buf, '\n') == NULL)
		;
	buf[0] = '\0';
	return 1;
}

int isidentifier(const char *s)
{
	int i;

	if (!isalpha(s[0]) && s[0] != '_')
		return 0;
	for (i = 1; isalnum(s[i]) || s[i] == '_'; i++)
		;
	return i;
}

int isassignment(const char *s)
{
	int i;

	i = isidentifier(s);
	return i > 0 && s[i] == '=';
}

void getvars(word_t words[], int nwords)
{
	int i;

	for (i = 0; i < nwords; i++) {
		word_t *nw = &words[i];
		if (!isassignment(nw->w))
			break;
		nw->type = W_ASSGN;
	}
}

void assignvars(word_t words[], int nwords)
{
	int i;
	char *v;

	for (i = 0; i < nwords; i++) {
		word_t *nw = &words[i];
		if (nw->type != W_ASSGN)
			break;
		v = strchr(nw->w, '=');
		*v = '\0';
		setenv(nw->w, v + 1, 1);
	}
}

int expandinternal(char c, struct obuf *o, int quote)
{
	int n;

	n = 1;
	switch (c) {
	case '$':
		oputf(o, quote, "%d", shellpid);
		break;
	case '?':
		oputf(o, quote, "%d", lastresult);
		break;
	default:
		n = 0;
		break;
	}
	return n;
}

int expandvar(const char *src, struct obuf *o, int quote)
{
	char vname[MAX_LINE];
	char *value;
	int i;

	i = expandinternal(src[0], o, quote);
	if (i)
		return i;

	i = isidentifier(src);
	if (!i)
		return 0;
	strncpy(vname, src, i);
	vname[i] = '\0';
	value = getenv(vname);
	if (value == NULL)
		value = "";
	oputs(o, value, quote);
	return i;
}

int issep(char c, char *sep)
{
	return strchr(sep, c) != NULL;
}

int issepspace(char c, char *sep)
{
	return issep(c, sep) && isspace(c);
}

int issepnonspace(char c, char *sep)
{
	return issep(c, sep) && !isspace(c);
}

int fieldsplit(char *buf, int len, char *fields[], int maxfields)
{
	char *sep;
	int i, nfields;

	sep = getenv("IFS");
	if (!sep)
		sep = " \t\n";

	i = 0;
	nfields = 0;
	while (i < len && issepspace(buf[i], sep))
		buf[i++] = '\0';
	for (;;) {
		if (i >= len)
			break;
		if (nfields >= maxfields) {
			fprintf(stderr, "too many fields\n");
			error = 1;
			return 0;
		}
		fields[nfields++] = &buf[i];
		while (i < len && !issep(buf[i], sep))
			i++;
		if (i >= len)
			break;
		if (!issepnonspace(buf[i], sep)) {
			do {
				buf[i++] = '\0';
			} while (i < len && issepspace(buf[i], sep));
		}
		if (i >= len)
			break;
		if (issepnonspace(buf[i], sep)) {
			buf[i++] = '\0';
			while (i < len && issepspace(buf[i], sep))
				buf[i++] = '\0';
		}
	}
	return nfields;
}

int expandwordfields(char *src, int split, struct obuf *o, char *fields[], int maxfields)
{
	char *last;
	int i, nfields;
	char c;

	i = 0;
	nfields = 0;
	last = NULL;
	while ( (c = src[i]) != '\0') {
		int quote = 0;
		if (c == '"') {
			quote = QUOTE;
			c = '$';
		}
		if (c == '$') {
			int k, old;
			char *tmp[MAX_WORDS], *buf;

			buf = o->buf;
			old = o->avail;
			if ( (k = expandvar(&src[i + 1], o, quote)) ) {
				int f;

				if (o->avail < 0) {
					error = 1;
					return 0;
				}
				i += k + 1;
				if (old == o->avail || !split)
					continue;
				f = fieldsplit(buf,
						old - o->avail,
						tmp,
						maxfields - nfields);
				if (error)
					return 0;
				if (!f) {
					if (last != NULL) {
						fields[nfields++] = last;
						last = NULL;
					}
				} else {
					int j;

					if (last != NULL)
						tmp[0] = last;
					for (j = 0; j < f - 1; j++)
						fields[nfields++] = tmp[j];
					if (o->buf[-1] == '\0') {
						fields[nfields++] = tmp[f - 1];
						last = NULL;
					} else {
						last = tmp[f - 1];
					}
				}
				continue;
			}
		}
		if (last == NULL)
			last = o->buf;
		oput(o, c);
		i++;
	}
	oput(o, '\0');
	if (last != NULL)
		fields[nfields++] = last;
	return nfields;
}

void updatewords(char *buf, char *tmpbuf, int nbcount,
		word_t *words, word_t *tmpwrd, int nwcount)
{
	int i;

	memcpy(buf, tmpbuf, nbcount);
	for (i = 0; i < nwcount; i++) {
		word_t *o = &words[i], *n = &tmpwrd[i];
		o->w = buf + (n->w - tmpbuf);
		o->type = n->type;
		o->quot = n->quot;
	}
}

int expandall(char *buf, word_t words[], int nwords)
{
	char	tmpbuf[MAX_LINE];
	word_t	tmpwrd[MAX_WORDS];
	struct obuf o;
	int	i, nwcount;
	char	*fields[MAX_WORDS];

	oinit(&o, tmpbuf);

	nwcount = 0;
	for (i = 0; i < nwords; i++) {
		word_t	*cur_org = &words[i],
			*cur_tmp = &tmpwrd[nwcount];
		int	nfields;

		cur_tmp->w = o.buf;
		cur_tmp->type = cur_org->type;
		cur_tmp->quot = cur_org->quot;
		nfields  = expandwordfields(cur_org->w,
				cur_tmp->type == W_NORML,
				&o,
				fields,
				MAX_WORDS - nwcount);
		if (o.avail < 0) {
			fprintf(stderr, "line too long\n");
			error = 1;
			return 0;
		}
		if (!nfields && cur_tmp->quot)
			nwcount++;
		else {
			int k;

			for (k = 0; k < nfields; k++) {
				cur_tmp = &tmpwrd[nwcount++];
				cur_tmp->w = fields[k];
				cur_tmp->type = cur_org->type;
				cur_tmp->quot = cur_org->quot;
			}
		}
	}
	updatewords(buf, tmpbuf, MAX_LINE - o.avail, words, tmpwrd, nwcount);
	return nwcount;
}

void unquote(char *buf)
{
	int i;

	for (i = 0; buf[i] != '\0'; i++)
		buf[i] &= ~QUOTE;
}

void unquoteall(word_t argv[], int nwords)
{
	int i;
	for (i = 0; i < nwords; i++)
		unquote(argv[i].w);
}

int getcont(char *buf)
{
	if (feof(stdin))
		return 0;
	if (interactive) {
		fprintf(stderr, "> ");
	}
	return fgets(buf, MAX_LINE, stdin) != NULL;
}

void quote(char *buf)
{
	char tmp[MAX_LINE];
	struct obuf o;
	int i;

	oinit(&o, tmp);

	i = 0;
	while (buf[i] != '\0') {
		switch (buf[i]) {
		case '\'':
			i++;
			while (buf[i] != '\'') {
				if (buf[i] == '\0') {
					if (getcont(buf)) {
						i = 0;
						continue;
					}
					fprintf(stderr, "unterminated single quote\n");
					error = 1;
					goto out;
				}
				oput(&o, buf[i++] | QUOTE);
			}
			i++;
			oput(&o, '\'');
			break;
		case '"':
			i++;
			while (buf[i] != '"') {
				if (buf[i] == '\0') {
					if (getcont(buf)) {
						i = 0;
						continue;
					}
					fprintf(stderr, "unterminated double quote\n");
					error = 1;
					goto out;
				}
				if (buf[i] == '\\' && (
						buf[i + 1] == '$' ||
						buf[i + 1] == '"' ||
						buf[i + 1] == '\\'
					))
				{
					i++;
				} else if (buf[i] == '\\' && buf[i + 1] == '\n') {
					i += 2;
					continue;
				} else if (buf[i] == '$') {
					oput(&o, '"');
					oput(&o, buf[++i]);
				       	if (buf[i] != '_' && !isalpha(buf[i])) {
						i++;
						continue;
					}
					for (i++; buf[i] == '_' || isalnum(buf[i]); i++)
						oput(&o, buf[i]);
					continue;
				}
				oput(&o, buf[i++] | QUOTE);
			}
			i++;
			oput(&o, '\'');
			break;
		case '\\':
			i++;
			if (buf[i] == '\n')
				i++;
			else
				oput(&o, '\'');
			if (buf[i] == '\0') {
				if (getcont(buf)) {
					i = 0;
					break;
				}
				goto out;
			}
			oput(&o, buf[i++] | QUOTE);
			break;
		default:
			oput(&o, buf[i++]);
			break;
		}
	}
out:
	oput(&o, '\0');
	if (o.avail < 0) {
		fprintf(stderr, "line too long\n");
		error = 1;
		tmp[0] = '\0';
	}
	strcpy(buf, tmp);
}

builtin_t findbuiltin(const char *name)
{
	builtin_des *d;

	if (name == NULL)
		return builtin_colon;

	for (d = builtins; d->name != NULL; d++) {
		if (!strcmp(name, d->name))
			return d->builtin;
	}
	return NULL;
}

int builtin_cd(int argc, char *argv[])
{
	char *dir;

	dir = argv[1];
	if (dir == NULL)
		dir = getenv("HOME");
	if (dir == NULL)
		return 0;
	if (chdir(dir) < 0) {
		perror(dir);
		error = 1;
		return 1;
	}
	return 0;

}

int builtin_umask(int argc, char *argv[])
{
	mode_t m;

	if (argc > 2) {
		fprintf(stderr, "syntax error\n");
		error = 1;
		return 1;
	}
	m = umask(0);
	if (argc == 1) {
		printf("0%03o\n", m);
	} else {
		m = strtol(argv[1], NULL, 8);
	}
	umask(m);
	return 0;
}

int builtin_unset(int argc, char *argv[])
{
	int i;

	for (i = 1; i < argc; i++)
		unsetenv(argv[i]);

	return 0;
}

int builtin_colon(int argc, char *argv[])
{
	return 0;
}

