#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* We add the `&&` , `||` and `!` logic operators available since V7.
 * These allow for simple if-then logic, but are easier to parse.
 */

#define MAX_LINE 1024
#define MAX_WORDS 100
#define MAX_ARGS  100

struct script {
	int interactive;
	int error;
	sig_t inth, quith;
	int lastresult;
	int lastbgpid;
	FILE *input;
	char *name;
	char **argv;
	int argc;
	char argbuf[MAX_LINE];
};

pid_t shellpid;

int interp(struct script *s);

int main(int argc, char *argv[])
{
	char *o, *name;
	FILE *input;
	int i, opt_i, opt_c, opt_s, rv;
	struct script s;

	setenv("IFS", " \t\n", 1);

	setbuf(stdin, NULL);
	shellpid = getpid();

	dup2(0, 3);
	dup2(1, 4);

	opt_i = 0;
	opt_c = 0;
	opt_s = 0;
	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-')
			break;
		for (o = &argv[i][1]; *o != '\0'; o++) {
			switch (*o) {
			case 'c':
				opt_c = 1;
				break;
			case 'i':
				opt_i = 1;
				break;
			case 's':
				opt_s = 1;
				break;
			default:
				fprintf(stderr, "unknown option: %c\n", *o);
				exit(1);
			}
		}
	}
	name = argv[0];
	argv += i - 1;
	argc -= i - 1;
	if (opt_c) {
		if (argc < 2) {
			fprintf(stderr, "missing command after -c\n");
			exit(1);
		}
		if ( (input = fmemopen(argv[1], strlen(argv[1]), "r")) == NULL ) {
			fprintf(stderr, "internal error\n");
			exit(1);
		}
		argc -= 2;
		argv += 2;
		if (argc)
			name = argv[0];
		s.interactive = 0;
	} else if (argc > 1 && !opt_s) {
		argc--;
		argv++;
		name = argv[0];
		if ( (input = fopen(name, "r")) == NULL ) {
			perror(name);
			exit(errno == ENOENT ? 127 : 126);
		}
		s.interactive = 0;
	} else {
		input = stdin;
		s.interactive = (isatty(0) && isatty(2)) || opt_i;
	}

	s.error = 0;
	s.inth = SIG_DFL;
	s.quith = SIG_DFL;
	if (s.interactive) {
		s.inth = signal(SIGINT, SIG_IGN);
		s.quith = signal(SIGQUIT, SIG_IGN);
	}
	s.lastresult = 0;
	s.lastbgpid = 0;
	s.input = input;
	s.name = name;
	s.argv = argv;
	s.argc = argc;

	return interp(&s);
}

typedef struct {
	char *w;
	int  type;
#define W_NORML	0
#define	W_REDIR	1
#define	W_ASSGN	2
	int quot;
} word_t;

#define QUOTE	0x80

struct obuf;
int getwords(struct script *s, char *buf, word_t words[]);
int buildargv(struct script *s, char *argv[], word_t words[], int nwords);
void getredirs(struct script *s, word_t words[], int nwords);
int redirect(struct script *s, word_t words[], int nwords, int in, int out);
int getquotedcmd(struct script *s, struct obuf *o);
void getvars(struct script *s, word_t words[], int nwords);
void assignvars(struct script *s, word_t words[], int nwords);
int expandall(struct script *s, char *buf, word_t words[], int nwords);
void unquoteall(struct script *s, word_t words[], int nwords);
void docmd(struct script *s, char *buf, char *term, int in, int out);
char *getsubcmd(struct script *s, char *line, char *buf, char *term);
void undoredirect(struct script *s);
int globall(struct script *s, char *buf, word_t words[], int nwords);
void dopipeline(struct script *s, char *buf, char *term);
void dolist(struct script *s, char *buf, char *term);

typedef int (*builtin_t)(struct script *s, int argc, char *argv[]);

typedef struct {
	const char *name;
	builtin_t  builtin;
} builtin_des;

int builtin_cd(struct script *s, int argc, char *argv[]);
int builtin_umask(struct script *s, int argc, char *argv[]);
int builtin_unset(struct script *s, int argc, char *argv[]);
int builtin_colon(struct script *s, int argc, char *argv[]);
int builtin_eval(struct script *s, int argc, char *argv[]);
int builtin_shift(struct script *s, int argc, char *argv[]);
int builtin_set(struct script *s, int argc, char *argv[]);
int builtin_dot(struct script *s, int argc, char *argv[]);
int builtin_exec(struct script *s, int argc, char *argv[]);
int builtin_exit(struct script *s, int argc, char *argv[]);
builtin_des builtins[] = {
	"cd",		builtin_cd,
	"umask",	builtin_umask,
	"unset",	builtin_unset,
	":",		builtin_colon,
	"eval",		builtin_eval,
	"shift",	builtin_shift,
	"set",		builtin_set,
	".",		builtin_dot,
	"exec",		builtin_exec,
	"exit", 	builtin_exit,
	NULL, NULL
};
builtin_t findbuiltin(struct script *s, const char *name);

struct obuf {
	char *buf;
	int avail;
	int size;
};

#define oinit(o, b) do {			\
		typeof(o) o__ = o;		\
		(o__)->buf = b;			\
		(o__)->avail = sizeof(b);	\
		(o__)->size = sizeof(b);	\
	} while(0)

int oused(struct obuf *o)
{
	if (o->avail < 0)
		return o->size;
	return o->size - o->avail;
}

void oput(struct obuf *o, char c)
{
	if (o->avail <= 0) {
		o->avail = -1;
		return;
	}
	*o->buf++ = c;
	o->avail--;
}

void oputs(struct obuf *o, const char *s, int quote)
{
	while (*s != '\0')
		oput(o, *s++ ^ quote);
}

void oputf(struct obuf *o, int quote, const char *fmt, ...)
{
	char tmp[MAX_LINE];
	va_list va;
	int w;

	va_start(va, fmt);
	w = vsnprintf(tmp, MAX_LINE, fmt, va);
	va_end(va);
	if (w >= 0)
		oputs(o, tmp, quote);
}

void odel(struct obuf *o)
{
	if (o->avail >= 0 && oused(o) > 0) {
		o->avail++;
		o->buf--;
	}
}

int interp(struct script *s)
{
	char buf[MAX_LINE];
	char line[MAX_LINE];
	char *term;
	struct obuf o;

	oinit(&o, line);
	while (getquotedcmd(s, &o)) {
		for (term = getsubcmd(s, line, buf, NULL);
		     term != NULL;
		     term = getsubcmd(s, line, buf, term))
		{
			dolist(s, buf, term);
		}
		oinit(&o, line);
	}
	return s->error ? s->error : s->lastresult;
}

void docmd(struct script *s, char *buf, char *term, int in, int out)
{
	word_t words[MAX_WORDS];
	char *c_argv[MAX_ARGS + 1];
	int nwords;
	int c_argc;
	builtin_t builtin;
	pid_t pid;

	nwords = getwords(s, buf, words);
	if (!nwords || s->error)
		return;
	getredirs(s, words, nwords);
	getvars(s, words, nwords);
	nwords = expandall(s, buf, words, nwords);
	nwords = globall(s, buf, words, nwords);
	c_argc = buildargv(s, c_argv, words, nwords);
	builtin = findbuiltin(s, c_argv[0]);
	unquoteall(s, words, nwords);

	if (builtin && *term != '&' && *term != '|') {
		assignvars(s, words, nwords);
		if (redirect(s, words, nwords, in, out) >= 0) {
			s->lastresult = (*builtin)(s, c_argc, c_argv);
		} else {
			s->lastresult = 1;
			s->error = 1;
		}
		if (builtin != builtin_exec)
			undoredirect(s);
		return;
	}

	if ( (pid = fork()) ) {
		int status;

		if (*term == '|')
			return;

		if (*term == '&') {
			s->lastbgpid = pid;
			s->lastresult = 0;
			return;
		}

		while (wait(&status) != pid)
			;
		if (WIFEXITED(status)) {
			s->lastresult = WEXITSTATUS(status);
		} else if (WIFSIGNALED(status)) {
			fprintf(stderr, "%s\n", strsignal(WTERMSIG(status)));
			s->lastresult = 0x80 | WTERMSIG(status);
		}
	} else {
		if (*term != '&') {
			signal(SIGINT, s->inth);
			signal(SIGQUIT, s->quith);
		} else if (s->interactive) {
			close(0);
			open("/dev/null", O_RDONLY);
		}
		if (redirect(s, words, nwords, in, out) < 0)
			exit(1);
		if (c_argv[0] == NULL)
			exit(0);
		assignvars(s, words, nwords);
		if (builtin)
			exit((*builtin)(s, c_argc, c_argv));
		execvp(c_argv[0], c_argv);
		perror(c_argv[0]);
		exit(errno == ENOENT ? 127 : 126);
	}
}

int cleanupword(char *w)
{
	int i, j, q;

	q = 0;
	for (i = 0, j = 0; w[i] != '\0'; i++) {
		if (w[i] != '\'')
			w[j++] = w[i];
		else
			q = 1;
	}
	w[j] = '\0';
	return q;
}

void cleanupall(word_t words[], int nwords)
{
	int i;

	for (i = 0; i < nwords; i++) {
		word_t *nw = &words[i];

		nw->quot = cleanupword(nw->w);
	}
}

int getwords(struct script *s, char *buf, word_t words[])
{
	int nwords;
	int i;
	word_t *nw;

	nwords = 0;
	i = 0;
	for (;;) {
		while (isspace(buf[i]))
			i++;
		if (buf[i] == '\0')
			break;
		if (nwords >= MAX_WORDS) {
			fprintf(stderr, "too many words\n");
			s->error = 1;
			nwords = 0;
			break;
		}
		nw = &words[nwords];
		nw->w = &buf[i];
		nw->type = W_NORML;
		nwords++;
		while (buf[i] != '\0' && !isspace(buf[i]))
			i++;
		if (buf[i] == '\0')
			break;
		buf[i++] = '\0';
	}
	cleanupall(words, nwords);
	return nwords;
}

int buildargv(struct script *s, char *argv[], word_t words[], int nwords)
{
	int i, argc;

	argc = 0;
	for (i = 0; i < nwords && i < MAX_ARGS; i++) {
		word_t *nw = &words[i];
		if (nw->type != W_NORML)
			continue;
		argv[argc] = nw->w;
		argc++;
	}
	if (argc >= MAX_ARGS) {
		fprintf(stderr, "too many arguments\n");
		s->error = 1;
		return 0;
	}
	argv[argc] = NULL;
	return argc;
}

void getredirs(struct script *s, word_t words[], int nwords)
{
	int i;

	for (i = 0; i < nwords; i++) {
		word_t *nw = &words[i];
		if (nw->w[0] == '<' || nw->w[0] == '>') {
			nw->type = W_REDIR;
			if (nw->w[0] == '>' && nw->w[1] == '>')
				nw->w[0] = 'A';
		}
	}
}

int redirect(struct script *s, word_t words[], int nwords, int in, int out)
{
	int i, flags;
	char *fname;
	word_t *nw;

	if (in != 0) {
		if (dup2(in, 0) < 0) {
			perror("dup2");
			return -1;
		}
		close(in);
	}
	if (out != 1) {
		if (dup2(out, 1) < 0) {
			perror("dup2");
			return -1;
		}
		close(out);
	}

	for (i = 0; i < nwords; i++) {
		nw = &words[i];
		if (nw->type != W_REDIR)
			continue;
		fname = &nw->w[1];
		if (nw->w[0] == '<') {
			flags = O_RDONLY;
			close(0);
		} else if (nw->w[0] == 'A') {
			flags = O_WRONLY | O_CREAT | O_APPEND;
			fname++;
			close(1);
		} else {
			flags = O_WRONLY | O_CREAT | O_TRUNC;
			close(1);
		}
		if (open(fname, flags, 0666) < 0) {
			perror(fname);
			return -1;
		}
	}
	return 0;
}

int getcmd(struct script *s, char *buf)
{
	if (s->interactive) {
		s->error = 0;
		clearerr(s->input);
		fprintf(stderr, geteuid() ? "$ " : "# ");
	}
	if (s->error)
		return 0;
	if (fgets(buf, MAX_LINE, s->input) == NULL)
		return 0;
	if (strchr(buf, '\n') != NULL || feof(s->input))
		return 1;
	fprintf(stderr, "line too long\n");
	if (!s->interactive) {
		s->error = 1;
		return 0;
	}
	while (fgets(buf, MAX_LINE, s->input) != NULL &&
		strchr(buf, '\n') == NULL)
		;
	buf[0] = '\0';
	return 1;
}

int isidentifier(const char *s)
{
	int i;

	if (!isalpha(s[0]) && s[0] != '_')
		return 0;
	for (i = 1; isalnum(s[i]) || s[i] == '_'; i++)
		;
	return i;
}

int isassignment(const char *s)
{
	int i;

	i = isidentifier(s);
	return i > 0 && s[i] == '=';
}

void getvars(struct script *s, word_t words[], int nwords)
{
	int i;

	for (i = 0; i < nwords; i++) {
		word_t *nw = &words[i];
		if (!isassignment(nw->w))
			break;
		nw->type = W_ASSGN;
	}
}

void assignvars(struct script *s, word_t words[], int nwords)
{
	int i;
	char *v;

	for (i = 0; i < nwords; i++) {
		word_t *nw = &words[i];
		if (nw->type != W_ASSGN)
			break;
		v = strchr(nw->w, '=');
		*v = '\0';
		setenv(nw->w, v + 1, 1);
	}
}

void concat(struct obuf *o, int n, char *argv[], int quote, int sep)
{
	int i;

	for (i = 0; i < n; i++) {
		if (i && (sep || !quote))
			oput(o, sep ^ quote);
		oputs(o, argv[i], quote);
	}
}

int expandinternal(struct script *s, char c, struct obuf *o, int quote)
{
	int n, a;
	char *ifs;
	int quotesep, sep;

	n = 1;
	quotesep = 0;
	switch (c) {
	case '$':
		oputf(o, quote, "%d", shellpid);
		break;
	case '?':
		oputf(o, quote, "%d", s->lastresult);
		break;
	case '!':
		oputf(o, quote, "%d", s->lastbgpid);
		break;
	case '#':
		oputf(o, quote, "%d", s->argc - 1);
		break;
	case '@':
		quotesep = quote;
		/* fallthrough */
	case '*':
		ifs = getenv("IFS");
		if (!ifs)
			sep = ' ';
		else
			sep = ifs[0];
		concat(o, s->argc - 1, &s->argv[1], quote, sep | quotesep);
		break;
	case '0':
		oputf(o, quote, "%s", s->name);
		break;
	default:
		if (c >= '1' && c <= '9') {
			a = c - '0';
			oputf(o, quote, "%s", (a < s->argc ? s->argv[a] : ""));
		} else {
			n = 0;
		}
		break;
	}
	return n;
}

int expandvar(struct script *s, const char *src, struct obuf *o, int quote)
{
	char vname[MAX_LINE];
	char *value;
	int i;

	i = expandinternal(s, src[0], o, quote);
	if (i)
		return i;

	i = isidentifier(src);
	if (!i)
		return 0;
	strncpy(vname, src, i);
	vname[i] = '\0';
	value = getenv(vname);
	if (value == NULL)
		value = "";
	oputs(o, value, quote);
	return i;
}

int issep(char c, char *sep)
{
	return strchr(sep, c) != NULL;
}

int issepspace(char c, char *sep)
{
	return issep(c, sep) && isspace(c);
}

int issepnonspace(char c, char *sep)
{
	return issep(c, sep) && !isspace(c);
}

int fieldsplit(struct script *s, char *buf, int len, char *fields[], int maxfields)
{
	char *sep;
	int i, nfields;

	sep = getenv("IFS");
	if (!sep)
		sep = " \t\n";

	i = 0;
	nfields = 0;
	while (i < len && issepspace(buf[i], sep))
		buf[i++] = '\0';
	for (;;) {
		if (i >= len)
			break;
		if (nfields >= maxfields) {
			fprintf(stderr, "too many fields\n");
			s->error = 1;
			return 0;
		}
		fields[nfields++] = &buf[i];
		while (i < len && !issep(buf[i], sep))
			i++;
		if (i >= len)
			break;
		if (!issepnonspace(buf[i], sep)) {
			do {
				buf[i++] = '\0';
			} while (i < len && issepspace(buf[i], sep));
		}
		if (i >= len)
			break;
		if (issepnonspace(buf[i], sep)) {
			buf[i++] = '\0';
			while (i < len && issepspace(buf[i], sep))
				buf[i++] = '\0';
		}
	}
	return nfields;
}

int expandwordfields(struct script *s, char *src, int split, struct obuf *o, char *fields[], int maxfields)
{
	char *last;
	int i, nfields;
	char c;

	i = 0;
	nfields = 0;
	last = NULL;
	while ( (c = src[i]) != '\0') {
		int quote = 0;
		if (c == '"') {
			quote = QUOTE;
			c = '$';
		}
		if (c == '$') {
			int k, old;
			char *tmp[MAX_WORDS], *buf;

			buf = o->buf;
			old = o->avail;
			if ( (k = expandvar(s, &src[i + 1], o, quote)) ) {
				int f;

				if (o->avail < 0) {
					s->error = 1;
					return 0;
				}
				i += k + 1;
				if (old == o->avail || !split)
					continue;
				f = fieldsplit(s, buf,
						old - o->avail,
						tmp,
						maxfields - nfields);
				if (s->error)
					return 0;
				if (!f) {
					if (last != NULL) {
						fields[nfields++] = last;
						last = NULL;
					}
				} else {
					int j;

					if (last != NULL)
						tmp[0] = last;
					for (j = 0; j < f - 1; j++)
						fields[nfields++] = tmp[j];
					if (o->buf[-1] == '\0') {
						fields[nfields++] = tmp[f - 1];
						last = NULL;
					} else {
						last = tmp[f - 1];
					}
				}
				continue;
			}
		}
		if (last == NULL)
			last = o->buf;
		oput(o, c);
		i++;
	}
	oput(o, '\0');
	if (last != NULL)
		fields[nfields++] = last;
	return nfields;
}

void updatewords(char *buf, char *tmpbuf, int nbcount,
		word_t *words, word_t *tmpwrd, int nwcount)
{
	int i;

	memcpy(buf, tmpbuf, nbcount);
	for (i = 0; i < nwcount; i++) {
		word_t *o = &words[i], *n = &tmpwrd[i];
		o->w = buf + (n->w - tmpbuf);
		o->type = n->type;
		o->quot = n->quot;
	}
}

int expandall(struct script *s, char *buf, word_t words[], int nwords)
{
	char	tmpbuf[MAX_LINE];
	word_t	tmpwrd[MAX_WORDS];
	struct obuf o;
	int	i, nwcount;
	char	*fields[MAX_WORDS];

	oinit(&o, tmpbuf);

	nwcount = 0;
	for (i = 0; i < nwords; i++) {
		word_t	*cur_org = &words[i],
			*cur_tmp = &tmpwrd[nwcount];
		int	nfields;

		cur_tmp->w = o.buf;
		cur_tmp->type = cur_org->type;
		cur_tmp->quot = cur_org->quot;
		nfields  = expandwordfields(s, cur_org->w,
				cur_tmp->type == W_NORML,
				&o,
				fields,
				MAX_WORDS - nwcount);
		if (o.avail < 0) {
			fprintf(stderr, "line too long\n");
			s->error = 1;
			return 0;
		}
		if (!nfields && cur_tmp->quot)
			nwcount++;
		else {
			int k;

			for (k = 0; k < nfields; k++) {
				cur_tmp = &tmpwrd[nwcount++];
				cur_tmp->w = fields[k];
				cur_tmp->type = cur_org->type;
				cur_tmp->quot = cur_org->quot;
			}
		}
	}
	updatewords(buf, tmpbuf, MAX_LINE - o.avail, words, tmpwrd, nwcount);
	return nwcount;
}

void unquote(char *buf)
{
	int i;

	for (i = 0; buf[i] != '\0'; i++)
		buf[i] &= ~QUOTE;
}

void unquoteall(struct script *s, word_t argv[], int nwords)
{
	int i;
	for (i = 0; i < nwords; i++)
		unquote(argv[i].w);
}

int getcont(struct script *s, char *buf)
{
	if (feof(stdin))
		return 0;
	if (s->interactive) {
		fprintf(stderr, "> ");
	}
	return fgets(buf, MAX_LINE, s->input) != NULL;
}

void quote(struct script *s, char *buf)
{
	char tmp[MAX_LINE];
	struct obuf o;
	int i;
	int skipquote;

	oinit(&o, tmp);

	i = 0;
	while (buf[i] != '\0') {
		switch (buf[i]) {
		case '\'':
			i++;
			while (buf[i] != '\'') {
				if (buf[i] == '\0') {
					if (getcont(s, buf)) {
						i = 0;
						continue;
					}
					fprintf(stderr, "unterminated single quote\n");
					s->error = 1;
					goto out;
				}
				oput(&o, buf[i++] | QUOTE);
			}
			i++;
			oput(&o, '\'');
			break;
		case '"':
			i++;
			skipquote = 0;
			while (buf[i] != '"') {
				if (buf[i] == '\0') {
					if (getcont(s, buf)) {
						i = 0;
						continue;
					}
					fprintf(stderr, "unterminated double quote\n");
					s->error = 1;
					goto out;
				}
				if (buf[i] == '\\' && (
						buf[i + 1] == '$' ||
						buf[i + 1] == '"' ||
						buf[i + 1] == '\\'
					))
				{
					i++;
				} else if (buf[i] == '\\' && buf[i + 1] == '\n') {
					i += 2;
					continue;
				} else if (buf[i] == '$') {
					oput(&o, '"');
					oput(&o, buf[++i]);
					if (buf[i] == '@')
						skipquote = 1;
				       	if (buf[i] != '_' && !isalpha(buf[i])) {
						i++;
						continue;
					}
					for (i++; buf[i] == '_' || isalnum(buf[i]); i++)
						oput(&o, buf[i]);
					continue;
				}
				oput(&o, buf[i++] | QUOTE);
			}
			i++;
			if (!skipquote)
				oput(&o, '\'');
			break;
		case '\\':
			i++;
			if (buf[i] == '\n')
				i++;
			else
				oput(&o, '\'');
			if (buf[i] == '\0') {
				if (getcont(s, buf)) {
					i = 0;
					break;
				}
				goto out;
			}
			oput(&o, buf[i++] | QUOTE);
			break;
		default:
			oput(&o, buf[i++]);
			break;
		}
	}
out:
	oput(&o, '\0');
	if (o.avail < 0) {
		fprintf(stderr, "line too long\n");
		s->error = 1;
		tmp[0] = '\0';
	}
	strcpy(buf, tmp);
}

builtin_t findbuiltin(struct script *s, const char *name)
{
	builtin_des *d;

	if (name == NULL)
		return builtin_colon;

	for (d = builtins; d->name != NULL; d++) {
		if (!strcmp(name, d->name))
			return d->builtin;
	}
	return NULL;
}

int builtin_cd(struct script *s, int argc, char *argv[])
{
	char *dir;

	dir = argv[1];
	if (dir == NULL)
		dir = getenv("HOME");
	if (dir == NULL)
		return 0;
	if (chdir(dir) < 0) {
		perror(dir);
		s->error = 1;
		return 1;
	}
	return 0;
}

int builtin_umask(struct script *s, int argc, char *argv[])
{
	mode_t m;

	if (argc > 2) {
		fprintf(stderr, "syntax error\n");
		s->error = 1;
		return 1;
	}
	m = umask(0);
	if (argc == 1) {
		printf("0%03o\n", m);
	} else {
		m = strtol(argv[1], NULL, 8);
	}
	umask(m);
	return 0;
}

int builtin_unset(struct script *s, int argc, char *argv[])
{
	int i;

	for (i = 1; i < argc; i++)
		unsetenv(argv[i]);

	return 0;
}

int builtin_colon(struct script *s, int argc, char *argv[])
{
	return 0;
}

int isterm(char c)
{
	return c == '\0' || c == '\n' || c == ';' || c == '&';
}

char *getsubcmd(struct script *s, char *line, char *buf, char *term)
{
	if (term == NULL) {
		if (line[0] == '\0')
			return NULL;
		term = line;
	} else {
		if (*term == '\0')
			return NULL;
		term++;
	}
again:
	while (!isterm(*term))
		*buf++ = *term++;
	if (*term == '&' && *(term + 1) == '&') {
		*buf++ = *term++;
		*buf++ = *term++;
		goto again;
	}
	*buf = '\0';
	return term;
}

void undoredirect(struct script *s)
{
	dup2(3, 0);
	dup2(4, 1);
}

int divert_interp(struct script *s, FILE *new_input)
{
	FILE *orig_input;
	int orig_interactive, rv;

	orig_input = s->input;
	orig_interactive = s->interactive;
	s->input = new_input;
	s->interactive = 0;
	rv = interp(s);
	s->input = orig_input;
	s->interactive = orig_interactive;
	return rv;
}

int builtin_eval(struct script *s, int argc, char *argv[])
{
	char buf[MAX_LINE];
	struct obuf o;
	FILE *e;
	int rv;

	oinit(&o, buf);

	if (argc < 2)
		return 0;
	concat(&o, argc - 1, &argv[1], 0, ' ');
	if (o.avail < 0) {
		s->error = 1;
		return 1;
	}
	if ( (e = fmemopen(buf, oused(&o), "r")) == NULL ) {
		fprintf(stderr, "out of memory\n");
		s->error = 1;
		return 1;
	}
	rv = divert_interp(s, e);
	fclose(e);
	return rv;
}

int builtin_shift(struct script *s, int argc, char *argv[])
{
	if (s->argc < 2)
		return 1;
	s->argv++;
	s->argc--;
	return 0;
}

int builtin_set(struct script *s, int argc, char *argv[])
{
	int w;
	struct obuf o;

	oinit(&o, s->argbuf);
	s->argc = 1;
	for (w = 1; w < argc; w++) {
		s->argv[s->argc++] = o.buf;
		oputs(&o, argv[w], 0);
		oput(&o, '\0');
	}
	if (o.avail < 0) {
		fprintf(stderr, "line too long\n");
		s->error = 1;
		s->argc = 1;
	}
	return 0;
}

int builtin_dot(struct script *s, int argc, char *argv[])
{
	char fullname[MAX_LINE];
	char *path;
	struct obuf o;
	int i, rv;
	FILE *input;

	if (argc < 2)
		return 0;

	if (strchr(argv[1], '/') == NULL) {
		path = getenv("PATH");
		if (path == NULL)
			path = "/bin:/usr/bin";
		fullname[0] = '\0';
		input = NULL;
		for (i = 0; path[i] != '\0'; i++) {
			oinit(&o, fullname);
			for (; path[i] != '\0' && path[i] != ':'; i++)
				oput(&o, path[i]);
			if (o.avail < MAX_LINE && o.buf[-1] != '/')
				oput(&o, '/');
			oputs(&o, argv[1], 0);
			oput(&o, '\0');
			if (o.avail < 0) {
				fprintf(stderr, "file name too long\n");
				s->error = 1;
				return 1;
			}
			if ( (input = fopen(fullname, "r")) != NULL )
				break;
		}
	} else {
		input = fopen(argv[1], "r");
	}
	if (input == NULL) {
		fprintf(stderr, "%s: not found\n", argv[1]);
		s->error = 1;
		return 1;
	}
	rv = divert_interp(s, input);
	fclose(input);
	return rv;
}

int builtin_exec(struct script *s, int argc, char *argv[])
{
	argc--;
	argv++;
	if (argc >= 1) {
		execvp(argv[0], argv);
		perror(argv[0]);
		s->error = 1;
		return 1;
	}
	return 0;
}

int match1(char* p, char* n)
{
	if (!*p)
		return !*n;

	if (*p == '*') {
		if (*n && match1(p, n + 1))
			return 1;
		return match1(p + 1, n);
	}

	if (!*n)
		return 0;

	if (*p == '?' || *p == *n)
		return match1(p + 1, n + 1);

	return 0;
}

int match(char* p, char* n)
{
	if (*n == '.') {
		if (*p != '.')
			return 0;
		else {
			p++;
			n++;
		}
	}
	return match1(p, n);
}

int ispattern(char *p)
{
	int i;

	for (i = 0; p[i] != '\0'; i++)
		if (p[i] == '?' || p[i] == '*')
			return 1;
	return 0;
}

void sort(char *fields[], int nfields)
{
	int i, j;
	char *tmp;

	for (i = 0; i < nfields; i++) {
		for (j = i + 1; j < nfields; j++) {
			if (strcmp(fields[i], fields[j]) > 0) {
				tmp = fields[i];
				fields[i] = fields[j];
				fields[j] = tmp;
			}
		}
	}
}

int glob1(struct script *s, char *p, char *w, struct obuf *o, char *fields[], int max)
{
	DIR *d;
	struct dirent* e;
	size_t l;
	int n;

	if (w[0] == '\0')
		return 0;

	if ( (d = opendir(p[0] == '\0' ? "." : p)) == NULL)
		return 0;
	n = 0;
	while ( (e = readdir(d)) ) {
		if (match(w, e->d_name)) {
			if (n >= max) {
				fprintf(stderr, "too many pathnames\n");
				s->error = 1;
				break;
			}
			fields[n++] = o->buf;
			oputs(o, p, 0);
			if (p[0] != '\0' && o->buf[-1] != '/')
				oput(o, '/');
			oputs(o, e->d_name, 0);
			oput(o, '\0');
		}
	}
	closedir(d);
	return n;
}

int globrec(struct script *s, char *p, char *w, struct obuf *o, char *fields[], int max)
{
	char sw[MAX_LINE];
	char *sf[MAX_ARGS];
	char sp[MAX_LINE];
	int i, j, n, nfields;

	for (i = 0; i < MAX_LINE && w[i] != '\0' && w[i] != '/'; i++)
		sw[i] = w[i];
	if (i == MAX_LINE) {
		fprintf(stderr, "pattern too long\n");
		s->error = 1;
		return 0;
	}
	sw[i] = '\0';
	n = glob1(s, p, sw, o, sf, max);
	if (w[i] == '/') {
		nfields = 0;
		for (j = 0; j < n; j++) {
			int k, q;

			for (k = 0; k < MAX_LINE && sf[j][k] != '\0'; k++)
				sp[k] = sf[j][k];
			for (q = i; k < MAX_LINE && w[q] == '/'; k++, q++)
				sp[k] = '/';
			if (k == MAX_LINE) {
				fprintf(stderr, "path too long\n");
				s->error = 1;
				return 0;
			}
			sp[k] = '\0';
			nfields += globrec(s, sp, &w[q], o, &fields[nfields], max - nfields);
		}
	} else {
		nfields = n;
		for (j = 0; j < n; j++)
			fields[j] = sf[j];
	}
	return nfields;
}

int glob(struct script *s, char *w, struct obuf *o, char *fields[], int max)
{
	int i, j, n;
	char p[MAX_LINE];

	if (!ispattern(w))
		return 0;

	j = 0;
	for (i = 0;  i < MAX_LINE && !strchr("*?", w[i]); i++)
		if ( (p[i] = w[i]) == '/' )
			j = i;
	if (i == MAX_LINE) {
		fprintf(stderr, "pattern too long\n");
		s->error = 1;
		return 0;
	}

	while (w[j] == '/') {
		p[j] = w[j];
		j++;
	}
	p[j] = '\0';
	n = globrec(s, p, &w[j], o, fields, max);
	sort(fields, n);
	return n;
}

int globall(struct script *s, char *buf, word_t words[], int nwords)
{
	char tmp[MAX_LINE];
	word_t new_words[MAX_WORDS];
	struct obuf o;
	int i, nnew;
	char *fields[MAX_WORDS];

	oinit(&o, tmp);

	nnew = 0;
	for (i = 0; i < nwords; i++) {
		word_t *cur_old = &words[i],
		       *cur_new = &new_words[nnew];
		int f;

		cur_new->w = o.buf;
		cur_new->type = cur_old->type;
		cur_new->quot = cur_old->quot;
		if (cur_new->type == W_NORML) {
			f  = glob(s, cur_old->w, &o, fields, MAX_WORDS - nnew);
			if (o.avail < 0) {
				fprintf(stderr, "line too long\n");
				s->error = 1;
				return 0;
			}
		} else {
			f = 0;
		}
		if (!f) {
			oputs(&o, cur_old->w, 0);
			oput(&o, '\0');
			nnew++;
		} else {
			int j;

			for (j = 0; j < f; j++) {
				cur_new = &new_words[nnew++];
				cur_new->w = fields[j];
				cur_new->type = cur_old->type;
				cur_new->quot = cur_old->quot;
			}
		}
	}
	if (o.avail < 0) {
		fprintf(stderr, "line too long\n");
		s->error = 1;
		return 0;
	}
	updatewords(buf, tmp, oused(&o), words, new_words, nnew);
	return nnew;
}

int needscont(char *line, int n)
{
	if (n <= 0)
		return 0;
	line--;
	while (n > 0 && isspace(*line)) {
		n--;
		line--;
	}
	return (n > 1 && *line == '|') ||
	       (n > 2 && *line == '&' && *(line - 1) == '&');
}

int getquotedcmd(struct script *s, struct obuf *o)
{
	char tmp[MAX_LINE];

	if (!getcmd(s, tmp))
		return 0;
	quote(s, tmp);
	oputs(o, tmp, 0);
	while (o->avail > 0 && needscont(o->buf, oused(o))) {
		if (!getcont(s, tmp))
			break;
		quote(s, tmp);
		odel(o);
		oputs(o, tmp, 0);
	}
	oput(o, '\0');
	return 1;
}

char *getsimplecmd(struct script *s, char *buf, char *cmd, char *next)
{
	if (next == NULL) {
		if (buf[0] == '\0')
			return NULL;
		next = buf;
	} else {
		if (*next == '\0')
			return NULL;
		next++;
	}
	while (*next != '\0' && *next != '|')
		*cmd++ = *next++;
	*cmd = '\0';
	return next;
}

void dopipeline(struct script *s, char *buf, char *term)
{
	char cmd[MAX_LINE];
	char *next;
	int p[2], last;

	last = 0;
	for (next = getsimplecmd(s, buf, cmd, NULL);
	     next != NULL;
	     next = getsimplecmd(s, buf, cmd, next))
	{
		if (*next == '|') {
			if (pipe(p) < 0) {
				perror("pipe");
				s->error = 1;
				return;
			}
			docmd(s, cmd, next, last, p[1]);
			close(p[1]);
			if (last)
				close(last);
			last = p[0];
		} else {
			docmd(s, cmd, term, last, 1);
			if (last)
				close(last);
			close(p[0]);
		}
	}
}

int ispipelineterm(char *next)
{
	char term;

	if (*next == '\0')
		return 1;
	term = *next++;
	if (term != '|' && term != '&')
		return 0;
	return term == *next;
}

char *getpipeline(struct script *s, char *buf, char *cmd, char *next)
{
	if (next == NULL) {
		if (buf[0] == '\0')
			return NULL;
		next = buf;
	} else {
		if (*next == '\0')
			return NULL;
		next += 2;
	}
	while (!ispipelineterm(next))
		*cmd++ = *next++;
	*cmd = '\0';
	return next;
}

int hasbang(char *cmd)
{
	int i;

	i = 0;
	while (isspace(cmd[i]))
		i++;
	if (cmd[i] == '!' && isspace(cmd[i + 1]))
		return i + 1;
	return 0;
}

void dolist(struct script *s, char *buf, char *term)
{
	char cmd[MAX_LINE];
	char *next;
	int i;

	for (next = getpipeline(s, buf, cmd, NULL);
	     next != NULL;
	     next = getpipeline(s, buf, cmd, next))
	{
		i = hasbang(cmd);
		dopipeline(s, &cmd[i], term);
		if (i)
			s->lastresult = !s->lastresult;
		if ((*next == '&' && s->lastresult) || (*next == '|' && !s->lastresult)) {
			next = getpipeline(s, buf, cmd, next);
			if (next == NULL)
				break;
		}
	}
}

int builtin_exit(struct script *s, int argc, char *argv[])
{
	int rv;

	rv = 0;
	if (argc > 1)
		rv = atoi(argv[1]);
	exit(rv);
}

